@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2020 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2023 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -39,11 +39,14 @@ import yaml
 import sys
 import subprocess
 import io
+import csv
 
 from enum import Enum
 from lxml import etree
 from lxml.cssselect import CSSSelector
 
+from xml.dom import minidom
+
 from .util import TrackSubClasses, import_module_from_source
 
 from .html2txt import html2text
@@ -79,6 +82,11 @@ try:
 except ImportError:
     Image = None
 
+try:
+    import jq
+except ImportError:
+    jq = None
+
 logger = logging.getLogger(__name__)
 
 
@@ -115,7 +123,8 @@ class FilterBase(object, metaclass=TrackSubClasses):
             filter_instance = filtercls(state.job, state)
             if filter_instance.match():
                 logger.info('Auto-applying filter %r to %s', filter_instance, state.job.get_location())
-                data = filter_instance.filter(data)
+                # filters require a subfilter argument
+                data = filter_instance.filter(data, None)
 
         return data
 
@@ -249,7 +258,10 @@ class LegacyHooksPyFilter(FilterBase):
     def match(self):
         return self.hooks is not None
 
-    def filter(self, data):
+    def filter(self, data, subfilter):
+        if subfilter is not None:
+            logger.warning('Legacy hooks filter does not have any subfilter -- ignored')
+
         try:
             result = self.hooks.filter(self.job.get_location(), data)
             if result is None:
@@ -309,8 +321,7 @@ class Html2TextFilter(FilterBase):
     def filter(self, data, subfilter):
         if 'method' in subfilter:
             method = subfilter['method']
-            del subfilter['method']
-            options = subfilter
+            options = {key: val for (key, val) in subfilter.items() if key != 'method'}
         else:
             method = 're'
             options = {}
@@ -319,6 +330,49 @@ class Html2TextFilter(FilterBase):
                          method=method, options=options)
 
 
+class Csv2TextFilter(FilterBase):
+    """Convert CSV to plaintext"""
+
+    __kind__ = 'csv2text'
+
+    __supported_subfilters__ = {
+        'format_message': 'A string used to format for each CSV line',
+        'ignore_header': 'Force index-based format string, even if headers are used',
+        'has_header': 'Override auto-detection of the first line as header',
+    }
+
+    __default_subfilter__ = 'format_message'
+
+    def filter(self, data, subfilter):
+        has_header_config = subfilter.get('has_header')
+
+        if has_header_config is None:
+            has_header = csv.Sniffer().has_header(data)
+        else:
+            has_header = has_header_config
+
+        reader = csv.reader(data.split('\n'))
+        data_list = list(reader)
+        header = None
+
+        if has_header:
+            header = data_list.pop(0)
+
+        header = [i.lower() for i in header]
+        message = subfilter['format_message']
+        ignore_header = subfilter['ignore_header']
+
+        lines = []
+        for i in data_list:
+            if header and not ignore_header:
+                legend = dict(zip(header, i))
+                lines.append(message.format(**legend))
+            else:
+                lines.append(message.format(*i))
+
+        return '\n'.join(lines)
+
+
 class Pdf2TextFilter(FilterBase):
     """Convert PDF to plaintext"""
     # Requires data to be in bytes (not unicode)
@@ -372,6 +426,22 @@ class JsonFormatFilter(FilterBase):
         return json.dumps(parsed_json, ensure_ascii=False, sort_keys=True, indent=indentation, separators=(',', ': '))
 
 
+class PrettyXMLFilter(FilterBase):
+    """Pretty-print XML"""
+
+    __kind__ = 'pretty-xml'
+
+    __supported_subfilters__ = {
+        'indentation': 'Indentation level for pretty-printing',
+    }
+
+    __default_subfilter__ = 'indentation'
+
+    def filter(self, data, subfilter):
+        indentation = int(subfilter.get('indentation', 2))
+        return minidom.parseString(data).toprettyxml(indent=' ' * indentation)
+
+
 class GrepFilter(FilterBase):
     """Filter only lines matching a regular expression"""
 
@@ -421,6 +491,17 @@ class StripFilter(FilterBase):
         return data.strip()
 
 
+class StripLinesFilter(FilterBase):
+    """Strip leading and trailing whitespace in every line"""
+
+    __kind__ = 'striplines'
+
+    __no_subfilter__ = True
+
+    def filter(self, data, subfilter=None):
+        return '\n'.join(line.strip() for line in data.splitlines())
+
+
 class FilterBy(Enum):
     ATTRIBUTE = 1
     TAG = 2
@@ -782,6 +863,31 @@ class SortFilter(FilterBase):
         return separator.join(sorted(data.split(separator), key=str.casefold, reverse=reverse))
 
 
+class RemoveDuplicateLinesFilter(FilterBase):
+    """Remove duplicate lines"""
+
+    __kind__ = 'remove-duplicate-lines'
+
+    __supported_subfilters__ = {
+        'separator': 'Item separator (default: newline)',
+    }
+
+    __default_subfilter__ = 'separator'
+
+    def filter(self, data, subfilter):
+        separator = subfilter.get('separator', '\n')
+        data_lines = data.split(separator)
+
+        def get_unique_lines(lines):
+            seen = set()
+            for line in lines:
+                if line not in seen:
+                    yield line
+                    seen.add(line)
+
+        return separator.join(get_unique_lines(data_lines))
+
+
 class ReverseFilter(FilterBase):
     """Reverse input items"""
 
@@ -851,3 +957,30 @@ class OCRFilter(FilterBase):
             raise ImportError('Please install Pillow/PIL')
 
         return pytesseract.image_to_string(Image.open(io.BytesIO(data)), lang=language, timeout=timeout)
+
+
+class JQFilter(FilterBase):
+    """Parse, transform, and extract data from json as text using `jq`"""
+
+    __kind__ = 'jq'
+
+    __supported_subfilters__ = {
+        'query': 'jq query function to execute on data',
+    }
+
+    __default_subfilter__ = 'query'
+
+    def filter(self, data, subfilter):
+
+        try:
+            jsondata = json.loads(data)
+        except ValueError:
+            raise ValueError('The url response contained invalid JSON')
+
+        if 'query' not in subfilter:
+            raise ValueError('{} filter needs a query'.format(self.__kind__))
+
+        if jq is None:
+            raise ImportError('Please install jq')
+
+        return jq.text(subfilter['query'], jsondata)
