@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2020 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2023 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -50,12 +50,23 @@ logger = logging.getLogger(__name__)
 class ShellError(Exception):
     """Exception for shell commands with non-zero exit code"""
 
-    def __init__(self, result):
+    def __init__(self, result, stdout_data, stderr_data, stderr_config):
         Exception.__init__(self)
         self.result = result
+        self.stdout_data = stdout_data
+        self.stderr_data = stderr_data
+        self.stderr_config = stderr_config
 
     def __str__(self):
-        return '%s: Exit status %d' % (self.__class__.__name__, self.result)
+        stdout = '\n'.join(('=' * 30, 'stdout from failed process:', repr(self.stdout_data), '=' * 30))
+        stderr = '\n'.join(('=' * 30, 'stderr from failed process:', repr(self.stderr_data), '=' * 30))
+        return '%s: Exit status %d\nstderr behavior: %r\n%s\n%s' % (
+            self.__class__.__name__,
+            self.result,
+            self.stderr_config,
+            stdout,
+            stderr,
+        )
 
 
 class NotModifiedError(Exception):
@@ -102,6 +113,9 @@ class JobBase(object, metaclass=TrackSubClasses):
     def get_location(self):
         raise NotImplementedError()
 
+    def set_base_location(self):
+        raise NotImplementedError()
+
     def pretty_name(self):
         raise NotImplementedError()
 
@@ -180,10 +194,11 @@ class JobBase(object, metaclass=TrackSubClasses):
 
 class Job(JobBase):
     __required__ = ()
-    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool', 'compared_versions', 'diff_filter')
+    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool', 'compared_versions', 'diff_filter', 'treat_new_as_changed', 'user_visible_url')
 
     # determine if hyperlink "a" tag is used in HtmlReporter
-    LOCATION_IS_URL = False
+    def location_is_url(self):
+        return re.match("^([a-zA-Z0-9+.-]+)://", self.get_location())
 
     def pretty_name(self):
         return self.name if self.name else self.get_location()
@@ -195,17 +210,39 @@ class ShellJob(Job):
     __kind__ = 'shell'
 
     __required__ = ('command',)
-    __optional__ = ()
+    __optional__ = ('stderr',)
 
     def get_location(self):
-        return self.command
+        return self.user_visible_url or self.command
+
+    def set_base_location(self, location):
+        self.command = location
 
     def retrieve(self, job_state):
-        process = subprocess.Popen(self.command, stdout=subprocess.PIPE, shell=True)
+        if not self.stderr or self.stderr == 'ignore':
+            # Report stderr output for non-zero exit code,
+            # but ignore stderr output with zero exit code
+            stderr = subprocess.PIPE
+        elif self.stderr == 'urlwatch':
+            # Legacy behavior, forward stderr output to urlwatch's stderr
+            stderr = None
+        elif self.stderr == 'fail':
+            # Treat any output on stderr as failure (even with zero exit code)
+            stderr = subprocess.PIPE
+        elif self.stderr == 'stdout':
+            # Combine stderr into stdout (kind of like "2>&1" on the shell)
+            stderr = subprocess.STDOUT
+        else:
+            raise ValueError('Invalid value for "stderr": %s' % (self.stderr,))
+
+        process = subprocess.Popen(self.command, stdout=subprocess.PIPE, stderr=stderr, shell=True)
         stdout_data, stderr_data = process.communicate()
         result = process.wait()
         if result != 0:
-            raise ShellError(result)
+            raise ShellError(result, stdout_data, stderr_data, self.stderr)
+        elif self.stderr == 'fail' and stderr_data:
+            # Exit code zero, but stderr contains data, and we want to fail
+            raise ShellError(result, stdout_data, stderr_data, self.stderr)
 
         if FilterBase.filter_chain_needs_bytes(self.filter):
             return stdout_data
@@ -223,11 +260,13 @@ class UrlJob(Job):
                     'headers', 'ignore_connection_errors', 'ignore_http_error_codes', 'encoding', 'timeout',
                     'ignore_timeout_errors', 'ignore_too_many_redirects')
 
-    LOCATION_IS_URL = True
     CHARSET_RE = re.compile('text/(html|plain); charset=([^;]*)')
 
     def get_location(self):
-        return self.url
+        return self.user_visible_url or self.url
+
+    def set_base_location(self, location):
+        self.url = location
 
     def retrieve(self, job_state):
         headers = {
@@ -251,12 +290,14 @@ class UrlJob(Job):
             headers['Cache-Control'] = 'max-age=172800'
             headers['Expires'] = email.utils.formatdate()
 
+        if self.data is not None:
+            if self.method is None:
+                self.method = "POST"
+            headers['Content-type'] = 'application/x-www-form-urlencoded'
+            logger.info('Sending %s request to %s', self.method, self.url)
+
         if self.method is None:
             self.method = "GET"
-        if self.data is not None:
-            self.method = "POST"
-            headers['Content-type'] = 'application/x-www-form-urlencoded'
-            logger.info('Sending POST request to %s', self.url)
 
         if self.http_proxy is not None:
             proxies['http'] = self.http_proxy
@@ -266,7 +307,8 @@ class UrlJob(Job):
         file_scheme = 'file://'
         if self.url.startswith(file_scheme):
             logger.info('Using local filesystem (%s URI scheme)', file_scheme)
-            return open(self.url[len(file_scheme):], 'rt').read()
+            with open(self.url[len(file_scheme):], 'rt') as f:
+                return f.read()
 
         if self.headers:
             self.add_custom_headers(headers)
@@ -364,10 +406,13 @@ class BrowserJob(Job):
 
     __required__ = ('navigate',)
 
-    LOCATION_IS_URL = True
+    __optional__ = ('wait_until', 'useragent')
 
     def get_location(self):
-        return self.navigate
+        return self.user_visible_url or self.navigate
+
+    def set_base_location(self, location):
+        self.navigate = location
 
     def main_thread_enter(self):
         from .browser import BrowserContext
@@ -377,4 +422,4 @@ class BrowserJob(Job):
         self.ctx.close()
 
     def retrieve(self, job_state):
-        return self.ctx.process(self.navigate)
+        return self.ctx.process(self.navigate, wait_until=self.wait_until, useragent=self.useragent)
